import marimo

__generated_with = "0.9.10"
app = marimo.App(width="medium", app_title="Benchmark", auto_download=["html"])


@app.cell
def __():
    import marimo as mo

    return (mo,)


@app.cell
def __(mo):
    mo.md("""# Benchmark""")
    return


@app.cell
def __():
    import time

    from multiprocessing import cpu_count

    import polars as pl
    import polars_simed as ps

    from faker import Faker

    import plotly.express as px

    return Faker, cpu_count, pl, ps, px, time


@app.cell
def __(mo):
    def get_arg(name: str, type, default):
        arg_val = mo.cli_args().get(name)
        if arg_val is None:
            return default
        else:
            return type(arg_val)

    return (get_arg,)


@app.cell
def __(get_arg):
    size_left = get_arg("size_left", int, 5000)
    size_right = get_arg("size_right", int, 10_000)
    size_left, size_right
    return size_left, size_right


@app.cell
def __(pl):
    def append_row(df: pl.DataFrame, data):
        """
        Append a row to a polars dataframe.
        """
        return pl.concat([df, pl.DataFrame(data)], how="vertical_relaxed")

    return (append_row,)


@app.cell
def __(Faker, size_left, size_right):
    fake = Faker()
    fake.seed_instance(4321)
    names_small = [fake.name() for _ in range(size_left)]
    names_big = [fake.name() for _ in range(size_right)]
    return fake, names_big, names_small


@app.cell
def __(names_big, names_small, pl):
    df_left = pl.DataFrame({"name": names_small})
    df_right = pl.DataFrame({"name": names_big})
    return df_left, df_right


@app.cell
def __(Iterable, append_row, pl, ps, time):
    def benchmark(
        df_left: pl.DataFrame,
        df_right: pl.DataFrame,
        argument_name: str,
        argument_values: Iterable,
        value_dtype,
        **kwargs,
    ) -> pl.DataFrame:
        if kwargs is None:
            kwargs = {}

        df_bench = pl.DataFrame(
            schema={
                "argument_name": pl.Utf8,
                "argument_value": value_dtype,
                "time": pl.Float32,
                "size_left": pl.Int32,
                "size_right": pl.Int32,
            }
        )

        for val in argument_values:
            kwargs.update(
                {
                    argument_name: val,
                }
            )

            start = time.time()
            ps.join_sim(df_left, df_right, **kwargs)
            end = time.time()

            elapsed_time = end - start

            df_bench = append_row(
                df_bench,
                [
                    {
                        "argument_name": argument_name,
                        "argument_value": val,
                        "time": elapsed_time,
                        "size_left": len(df_left),
                        "size_right": len(df_right),
                    }
                ],
            )

        return df_bench

    return (benchmark,)


@app.cell
def __(px):
    def plot_benchmark(df_bench):
        argument_name = df_bench["argument_name"].unique()[0]
        size_left = df_bench["size_left"].unique()[0]
        size_right = df_bench["size_right"].unique()[0]

        fig = px.bar(
            df_bench,
            x="argument_value",
            y="time",
            title=f"Execution time for different {argument_name} values.<br>Dataframe dimensions: ({size_left},{size_right}).",
        )

        return fig

    return (plot_benchmark,)


@app.cell
def __(cpu_count, df_left, df_right, pl):
    benchmarks = {
        "ntop": {
            "df_left": df_left,
            "df_right": df_right,
            "argument_name": "top_n",
            "argument_values": range(1, 100, 10),
            "value_dtype": pl.Int32,
        },
        "threads": {
            "df_left": df_left,
            "df_right": df_right,
            "argument_name": "threads",
            "argument_values": range(1, cpu_count() + 1),
            "value_dtype": pl.Int32,
        },
        "normalizatåion": {
            "df_left": df_left,
            "df_right": df_right,
            "argument_name": "normalization",
            "argument_values": ["l2", "count"],
            "value_dtype": pl.Utf8,
        },
        "threading_coordinate": {
            "df_left": df_left,
            "df_right": df_right,
            "argument_name": "threading_dimension",
            "argument_values": ["left", "right"],
            "value_dtype": pl.Utf8,
        },
    }
    return (benchmarks,)


@app.cell
def __(benchmark, benchmarks, plot_benchmark):
    figs = {}
    for name, kw in benchmarks.items():
        df_l = benchmark(on="name", **kw, threading_dimension="left")
        df_r = benchmark(on="name", **kw, threading_dimension="right")

        figs[f"{name}_left"] = plot_benchmark(df_l)
        figs[f"{name}_right"] = plot_benchmark(df_r)
    return df_l, df_r, figs, kw, name


@app.cell
def __(figs):
    figs
    return


@app.cell
def __():
    return


if __name__ == "__main__":
    app.run()



import pytest
import polars as pl
from polars.testing import assert_frame_equal
import polars_simed as ps


@pytest.mark.parametrize(
    "left, right, expected",
    [
        (
            pl.DataFrame({"s": ["aaa"]}),
            pl.DataFrame({"s": ["aaa"]}),
            pl.DataFrame({"sim": [1], "row": [0], "col": [0]}),
        ),
        (
            pl.DataFrame({"s": ["aaabb"]}),
            pl.DataFrame({"s": ["aaa"]}),
            pl.DataFrame({"sim": [1 / 3**0.5], "row": [0], "col": [0]}),
        ),
        # check for symmetriy
        (
            pl.DataFrame({"s": ["aaa"]}),
            pl.DataFrame({"s": ["aaabb"]}),
            pl.DataFrame({"sim": [1 / 3**0.5], "row": [0], "col": [0]}),
        ),
        (
            # one matching token
            # right has three tokens
            pl.DataFrame({"s": ["abc"]}),
            pl.DataFrame({"s": ["abcabc"]}),
            pl.DataFrame({"sim": [1 / 3**0.5], "row": [0], "col": [0]}),
        ),
        (
            # left has two rows
            # right has two rows
            pl.DataFrame({"s": ["abc", "def"]}),
            pl.DataFrame({"s": ["abc", "aaa"]}),
            pl.DataFrame({"sim": [1], "row": [0], "col": [0]}),
        ),
        (
            # left has three rows
            # right has two rows
            pl.DataFrame({"s": ["abc", "def", "aaabxy"]}),
            pl.DataFrame({"s": ["abc", "aaa"]}),
            pl.DataFrame({"sim": [1.0, 1 / 4**0.5], "row": [0, 2], "col": [0, 1]}),
        ),
    ],
)
def test_join_sim_basic(left, right, expected):
    kwargs_to_test = [
        {"threads": 2, "threading_dimension": "right"},
        # {"threads": 2, "threading_dimension": "right"},
        # {"threads" : None, "threading_dimension": "left"},
        # {"threads" : None, "threading_dimension": "right"},
        # {"threading_dimension": "auto"}
    ]

    for kw in kwargs_to_test:
        result = ps.join_sim(
            left,
            right,
            on="s",
            top_n=1,
            normalization="l2",
            add_mapping=True,
            add_similarity=True,
            suffix="_right",
            **kw,
        )
        assert isinstance(result, pl.DataFrame)
        assert set(result.columns) == {"s", "s_right", "sim", "row", "col"}
        assert_frame_equal(
            result.drop("s", "s_right"),
            expected,
            check_column_order=False,
            check_dtypes=False,
            atol=1e-31,
        )



from typing import Literal
import polars as pl
from polars_simed._polars_simed_lib import awesome_cossim


def normalize_string_col(df: pl.DataFrame, col: str) -> pl.DataFrame:
    return df.with_columns(
        pl.col(col)
        .str.replace_all("[^a-zA-Z0-9]", "")
        .str.to_lowercase()
        .alias(f"{col}_normalized"),
    )


def join_sim(
    left: pl.DataFrame,
    right: pl.DataFrame,
    *,
    on: str | None = None,
    left_on: str | None = None,
    right_on: str | None = None,
    top_n: int = 10,
    normalization: Literal["l2", "count"] = "l2",
    apply_word_normalization: bool = False,
    threads: int | None = None,
    suffix: str = "_right",
    add_mapping: bool = False,
    add_similarity: bool = True,
    threading_dimension: Literal["left", "right", "auto"] = "left",
) -> pl.DataFrame:
    """
    Compute the row-wise similarity between two DataFrames. Behaves similar to a
    left join or join_asof for time series data.

    The similarity is the inner product of the (normalized) three-gram representation
    of the strings.

    Parameters:

        left (pl.DataFrame):
            The left DataFrame.
        right (pl.DataFrame):
            The right DataFrame.

        left_on (str):
            The column name in the left DataFrame to be used for similarity computation.

        right_on (str):
            The column name in the right DataFrame to be used for similarity computation.

        top_n (int):
            The maximum number of similar items to return for each row.

        normalization (Literal["l2", "count]):
            If "l2", the vectors are l2-normalized and the similarity is the cosine
            similarity. If "count", the similarity is the number of common elements/tokens.
            Currently, the tokens are simply three-grams of alphanumeric characters.
            NOTE: If "count" is used, the underlying computations are significantly faster due to
            u16 integer arithmetic.

        apply_word_normalization (bool):
            Whether to apply word normalization to the columns before computing the similarity.
            The normalization is done by removing all non-alphanumeric characters and converting
            the string to lowercase. Decreases the pre-processing time but might lead to better
            results.
            Defaults to False.

        threads (int | None, optional):
            The number of threads to use for computation. Defaults to None (= number of physical cores).

        add_mapping (bool):
            Whether to add the row-col mapping of df_left and df_right to the result.
            Defaults to False.

        add_similarity (bool):
            Whether to add the similarity score to the result.
            Defaults to True.

        threading_dimension (Literal["left", "right", "auto"]):
            The dimension to parallelize.
            If "left", the left DataFrame is parallelized.
            If "right", the right DataFrame is parallelized.
            If "auto", we use left parallelization if the left DataFrame is significantly smaller
            than the right DataFrame.
            Defaults to "left".

    Returns:
        pl.DataFrame:
            A DataFrame containing the cosine similarity results of length
            less or equal to len(left) * top_n.
    """

    if on is not None:
        left_on = on
        right_on = on

    assert left_on in left.columns, f"{left_on} not in left DataFrame"
    assert right_on in right.columns, f"{right_on} not in right DataFrame"

    normalize: bool = True

    match threading_dimension:
        case "left":
            parallelize_left = True
        case "right":
            parallelize_left = False
        case "auto":
            if 100 * len(left) <= len(right):
                # left side is very small
                # hence it might be benefitial to parallelize the right side
                # comes at a cost of transposing the right slices thread-wise
                parallelize_left = False
            else:
                parallelize_left = True
        case _:
            raise ValueError(f"Threading {threading_dimension} not supported")

    match normalization:
        case "l2":
            normalize = True
        case "count":
            normalize = False
        case _:
            raise ValueError(f"Normalization {normalization} not supported")

    if apply_word_normalization:
        left = normalize_string_col(left, left_on)
        right = normalize_string_col(right, right_on)
        left_on = f"{left_on}_normalized"
        right_on = f"{right_on}_normalized"

    _map = awesome_cossim(
        left, right, left_on, right_on, top_n, threads, normalize, parallelize_left
    ).cast(
        {
            "row": pl.UInt32,
            "col": pl.UInt32,
            "sim": pl.Float32,
        }
    )

    return (
        left.with_row_index("row")
        .join(_map, on="row", how="left")
        .join(right.with_row_index("col"), on="col", suffix=suffix)
        .pipe(lambda df: df.drop("row", "col") if not add_mapping else df)
        .pipe(lambda df: df.drop("sim") if not add_similarity else df)
    )
